r"""The core database functionality."""

# Standard library
from typing import TypeAlias

# Third party
from sqlalchemy import URL as _URL
from sqlalchemy import Engine as _Engine
from sqlalchemy import create_engine
from sqlalchemy.orm import Session as _Session
from sqlalchemy.orm import sessionmaker

# Local
from .models import Base

Engine: TypeAlias = _Engine
Session: TypeAlias = _Session
SessionFactory: TypeAlias = sessionmaker[_Session]
URL: TypeAlias = str | _URL


def create_session_factory(
    url: URL,
    autoflush: bool = False,
    expire_on_commit: bool = False,
    create_database: bool = True,
    **engine_config,
) -> tuple[SessionFactory, Engine]:
    r"""Create the database session factory, which can produce database sessions.

    The database engine is bound to each session produced by the factory.

    Parameters
    ----------
    url : str or sqlalchemy.URL
        The SQLAlchemy database url.

    autoflush : bool, default False
        Automatically flush pending changes within the session to the database
        before executing new SQL statements.

    expire_on_commit : bool, default False
        If True make the connection between the models and the database expire after a
        transaction within a session has been committed and if False make the database models
        accessible after the commit.

    create_database : bool, default True
        If True the database table schema will be created if it does not exist.

    **engine_config : dict
        Additional keyword arguments passed to the :func:`sqlalchemy.create_engine` function.

    Returns
    -------
    session_factory : sqlalchemy.orm.Session
        The session factory that can produce new database sessions.

    engine : sqlalchemy.Engine
        The database engine that manages the database connections and is bound
        to `SessionFactory`.
    """

    engine = create_engine(url=url, **engine_config)
    session_factory = sessionmaker(
        bind=engine, autoflush=autoflush, expire_on_commit=expire_on_commit
    )

    if create_database:
        Base.metadata.create_all(bind=engine)

    return session_factory, engine



r"""The ID:s of components used for identification and session state management."""

# =====================================================================================
# Register form
# =====================================================================================

BP_REGISTER_FORM = 'bp-register-form'
BP_REGISTER_FORM_USERNAME_TEXT_INPUT = 'bp-register-form-username-text-input'
BP_REGISTER_FORM_DISPLAYNAME_TEXT_INPUT = 'bp-register-form-displayname-text-input'
BP_REGISTER_FORM_ALIASES_TEXT_INPUT = 'bp-register-form-aliases-text-input'
BP_REGISTER_FORM_SUBMIT_BUTTON = 'bp-register-form-submit-button'

# =====================================================================================
# Sign in form
# =====================================================================================

BP_SIGN_IN_FORM_ALIAS_TEXT_INPUT = 'bp-sign-in-form-alias-text-input'
BP_SIGN_IN_FORM_SUBMIT_BUTTON = 'bp-sign-in-form-submit-button'



r"""The models that represent tables in the database."""

# Standard library
from datetime import datetime
from typing import Optional

# Third party
from sqlalchemy import TIMESTAMP, ForeignKey, Index, func
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

# Local

SCHEMA: str | None = None


class Base(DeclarativeBase):
    r"""The base model all database models will inherit from.

    It defines the schema/user where the tables will be created in the
    database and the columns, which are common for all tables.

    Parameters
    ----------
    modified_at : datetime
        The timestamp at which the table record was latest modified (UTC).

    created_at : datetime
        The timestamp at which the table record was created (UTC).
    """

    __table_args__ = {'schema': SCHEMA}

    modified_at: Mapped[datetime] = mapped_column(
        TIMESTAMP(),
        server_default=func.current_timestamp(),
        onupdate=func.current_timestamp(),
    )
    created_at: Mapped[datetime] = mapped_column(
        TIMESTAMP(), server_default=func.current_timestamp()
    )


class User(Base):
    r"""The user table.

    Parameters
    ----------
    user_id : str
        The unique ID of the user. The primary key of the table.

    username : str
        The username of the user. It must be unique across all users.

    ad_username : Optional[str]
        The active directory username of the user is such exists.

    displayname : Optional[str]
        A descriptive name of the user that is easy to understand for a human.

    disabled : bool, default False
        If False the user is enabled and if True the user is disabled.
        A disabled user is not able to register credentials or sign in.

    disabled_timestamp : Optional[datetime]
        The timestamp when the user was disabled.

    emails : list[Email]
        The email addresses associated with the user.
    """

    __tablename__ = 'stp_user'

    user_id: Mapped[str] = mapped_column(primary_key=True)
    username: Mapped[str] = mapped_column(unique=True)
    ad_username: Mapped[Optional[str]]
    displayname: Mapped[Optional[str]]
    disabled: Mapped[bool] = mapped_column(default=False)
    disabled_timestamp: Mapped[Optional[datetime]] = mapped_column(TIMESTAMP())
    emails: Mapped[list['Email']] = relationship(back_populates='user')

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}(\n'
            f'    user_id={self.user_id},\n'
            f'    username={self.username},\n'
            f'    ad_username={self.username},\n'
            f'    displayname={self.displayname},\n'
            f'    disabled={self.disabled},\n'
            f'    disabled_timestamp={self.disabled},\n'
            f'    emails={len(self.emails)}\n)'
        )


Index(f'{User.__tablename__}_username_ix', User.username)
Index(f'{User.__tablename__}_ad_username_ix', User.ad_username)
Index(f'{User.__tablename__}_disabled_ix', User.disabled)


class Email(Base):
    r"""Email addresses of a user.

    Parameters
    ----------
    email_id : int
        The primary key of the table.

    user_id : str
        The unique ID of the user the email address belongs to.

    email : str
        An email address of a user. Must be unique.

    active : bool, default False
        If the email address is active or not.

    user : User
        The user object the email address belongs to.
    """

    __tablename__ = 'stp_email'

    email_id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    user_id: Mapped[str] = mapped_column(ForeignKey(User.user_id))
    email: Mapped[str] = mapped_column(unique=True)
    active: Mapped[bool] = mapped_column(default=True)
    user: Mapped['User'] = relationship(back_populates='emails')

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}(\n'
            f'    email_id={self.email_id},\n'
            f'    user_id={self.user_id},\n'
            f'    email={self.email},\n'
            f'    active={self.active},\n)'
        )


Index(f'{Email.__tablename__}_email_ix', Email.email)
Index(f'{Email.__tablename__}_user_id_ix', Email.user_id)



r"""Database operations on the user table."""

# Standard library
from typing import Sequence

# Third party
import pandas as pd
from sqlalchemy import and_, select

# Local
from .. import models
from ..core import Session
from ..schemas import user as schemas


def get_all_users(
    session: Session,
    skip: int = 0,
    limit: int | None = None,
    as_df: bool = False,
    index_col: str = 'user_id',
) -> pd.DataFrame | Sequence[models.User]:
    r"""Get all users from the database.

    Parameters
    ----------
    session : Session
        An active database session.

    skip : int, default 0
        The number of users to skip from the beginning of the result set.
        The users are order by their username.

    limit : int or None, default None
        Limit the result to desired number of users. If None all users are returned.

    as_df : bool, default False
        Return the users as a :class:`pandas.DataFrame` instead of a sequence of user models.

    index_col : str, default 'user_id'
        The column from the user table to use as the index column
        of the :class:`pandas.DataFrame` if `as_df` is True.

    Returns
    -------
    pandas.DataFrame or Sequence[streamlit_passwordless.db.models.User]
        The users from the database.
    """

    query = select(models.User).order_by(models.User.username).offset(skip).limit(limit)

    if as_df:
        return pd.read_sql_query(sql=query, con=session.bind, index_col=index_col)  # type: ignore
    else:
        return session.scalars(query).all()


def get_user_by_username(
    session: Session, username: str, disabled: bool = False
) -> models.User | None:
    r"""Get a user by its username.

    Parameters
    ----------
    session : Session
        An active database session.

    username : str
        The username to filter by.

    disabled : bool, default False
        True if filtering for a disabled user and False for an active user.

    Returns
    -------
    streamlit_passwordless.db.models.User or None
        The user with `username` or None if a user with `username` was not found.
    """

    query = select(models.User).where(
        and_(models.User.username == username, models.User.disabled == disabled)
    )
    return session.scalars(query).one_or_none()


def create_user(session: Session, user: schemas.UserCreate) -> models.User:
    r"""Create a new user.

    The user is added to the session and will be persisted in the
    database when the method `session.commit` is executed.

    Parameters
    ----------
    session : Session
        An active database session.

    user : schemas.UserCreate
        The user to crete.

    Returns
    -------
    streamlit_passwordless.db.models.User
        The database model of the created user.
    """

    db_user = models.User(**user.model_dump())
    session.add(db_user)

    return db_user



r"""The schemas for the user model."""

# Standard library
import uuid
from datetime import datetime

# Third party
from pydantic import Field, field_validator

# Local
from .core import SchemaBaseModel


class UserBase(SchemaBaseModel):
    r"""The base schema of a user.

    Parameters
    ----------
    user_id : str or None, default None
        The unique ID of the user, which serves as the primary key in the database.
        If None it will be generated as a uuid.
    """

    user_id: str | None = Field(default=None, validate_default=True)

    @field_validator('user_id')
    def generate_user_id(cls, v: str | None) -> str:
        r"""Generate a user ID if not supplied."""

        return str(uuid.uuid4()) if v is None else v


class UserCreate(UserBase):
    r"""The schema for creating a new user.

    Parameters
    ----------
    username : str
        The username of the user.

    ad_username : str or None, default None
        The active directory username of the user.

    displayname : str or None, default None
        A descriptive name of the user that is easy to understand for a human.

    disabled : bool, default False
        If False the user is enabled and if True the user is disabled.
        A disabled user is not able to register credentials or sign in.

    disabled_timestamp : datetime or None, default None
        The timestamp when the user was disabled.
    """

    username: str
    ad_username: str | None = None
    displayname: str | None = None
    disabled: bool = False
    disabled_timestamp: datetime | None = None



r"""The register-form component and its callback functions."""

# Standard library
import logging
from datetime import timedelta

# Third party
import streamlit as st

# Local
from streamlit_passwordless import database as db
from streamlit_passwordless import exceptions, models
from streamlit_passwordless.bitwarden_passwordless.client import BitwardenPasswordlessClient
from streamlit_passwordless.bitwarden_passwordless.frontend import register_button
from streamlit_passwordless.web import get_origin_header

from . import config, core, ids

logger = logging.getLogger(__name__)


@st.cache_data
def _create_user(
    username: str,
    user_id: str | None = None,
    displayname: str | None = None,
    aliases: str | None = None,
) -> tuple[models.User | None, str]:
    r"""Create a new user to register.

    Parameters
    ----------
    username : str
        The username.

    user_id : str or None, default None
        The unique ID of the user, which serves as the primary key in the database.
        If None it will be generated as a uuid.

    displayname : str or None, default None
        The optional displayname of the user.

    aliases : str or None, default None
        The optional aliases of the user as a semicolon separated string.

    Returns
    -------
    user : models.User or None
        The user to create. None is returned if an error occurred while creating the user.

    error_msg : str
        The error message if a user could not be created successfully. If no errors an empty
        string is returned.
    """

    error_msg = ''
    try:
        user = models.User(
            username=username, user_id=user_id, displayname=displayname, aliases=aliases
        )
    except exceptions.StreamlitPasswordlessError as e:
        error_msg = str(e)
        logger.error(error_msg)
        user = None
    else:
        logger.debug(f'Successfully created user: {user}')

    return user, error_msg


@st.cache_data(
    ttl=timedelta(minutes=2),
    show_spinner=False,
    hash_funcs={BitwardenPasswordlessClient: hash, models.User: hash},
)
def _create_register_token(
    client: BitwardenPasswordlessClient, user: models.User
) -> tuple[str, str]:
    r"""Create a register token to register a new passkey with the user's device.

    Parameters
    ----------
    client : BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless API.

    user : models.User
        The user to register.

    Returns
    -------
    register_token : str
        The register token. An empty string is returned if an error occurred.

    error_msg : str
        An error message if the register token could not be created.
        An empty string is returned if no error occurred.
    """

    error_msg = ''

    try:
        register_token = client.create_register_token(user=user)
    except exceptions.RegisterUserError as e:
        error_msg = str(e)
        register_token = ''

    return register_token, error_msg


def _validate_form(
    db_session: db.Session,
    client: BitwardenPasswordlessClient,
    origin: str,
    pre_authorized: bool = False,
) -> None:
    r"""Validate the input fields of the register form.

    Updates the following session state keys:
    - `config.SK_DB_USER` :
        Assigns the database user object. None is assigned if the user does
        not exist in the database.

    - `config.SK_REGISTER_FORM_IS_VALID` :
        True if the form validation passed and False otherwise.

    Parameters
    ----------
    db_session : streamlit_passwordless.database.Session
        An active database session.

    client : streamlit_passwordless.BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless backend API.

    origin : str
        The domain name of the application. Used to fetch existing passkey credentials of the
        the user for the application. An authenticated user with existing credentials to the
        application may create new credentials.

    pre_authorized : bool, default False
        If True require a user with the input username to exist in the database to allow
        the user to register a new passkey credential. If False omit this validation.
    """

    error_msg = ''

    username = st.session_state[ids.BP_REGISTER_FORM_USERNAME_TEXT_INPUT]
    if not username:
        error_msg = 'The username field is required!'
        st.error(error_msg, icon=config.ICON_ERROR)
        st.session_state[config.SK_REGISTER_FORM_IS_VALID] = False
        return

    user = db.get_user_by_username(session=db_session, username=username)

    if not user:
        st.session_state[config.SK_DB_USER] = None

        if pre_authorized:
            error_msg = (
                f'User {username} does not exist, but is required to exist '
                'to allow registration!'
            )
            logger.warning(error_msg)

    elif user:
        st.session_state[config.SK_DB_USER] = user

        if origin:
            try:
                credentials = client.get_credentials(user_id=user.user_id, origin=origin)
            except exceptions.StreamlitPasswordlessError as e:
                error_msg = f'Could not get passkey credentials for user {username}!\n{str(e)}'
                credentials = []
        else:
            credentials = []

        if (verified_user := st.session_state.get(config.SK_BP_VERIFIED_USER)) is not None:
            is_authenticated = True if verified_user.user_id == user.user_id else False
        else:
            is_authenticated = False

        if not error_msg and credentials and not is_authenticated:
            error_msg = (
                f'User {username} already exist! '
                'To add additional passkeys, please sign in first.'
            )
            logger.warning(error_msg)

    if error_msg:
        st.error(error_msg, icon=config.ICON_ERROR)
        st.session_state[config.SK_REGISTER_FORM_IS_VALID] = False
    else:
        st.session_state[config.SK_REGISTER_FORM_IS_VALID] = True


def bitwarden_register_form(
    client: BitwardenPasswordlessClient,
    db_session: db.Session,
    is_admin: bool = False,
    pre_authorized: bool = False,
    with_displayname: bool = False,
    with_alias: bool = False,
    title: str = '#### Register a new passkey with your device',
    border: bool = True,
    validate_button_label: str = 'Validate',
    register_button_label: str = 'Register',
    validate_button_type: core.ButtonType | None = None,
    register_button_type: core.ButtonType = 'primary',
    clear_on_validate: bool = False,
    username_label: str = 'Username',
    username_max_length: int | None = 50,
    username_placeholder: str | None = 'john.doe@example.com',
    username_help: str | None = '__default__',
    displayname_label: str = 'Displayname',
    displayname_max_length: int | None = 50,
    displayname_placeholder: str | None = 'John Doe',
    displayname_help: str | None = '__default__',
    alias_label: str = 'Alias',
    alias_max_length: int | None = 50,
    alias_placeholder: str | None = 'j;john;jd',
    alias_help: str | None = '__default__',
) -> None:
    r"""Render the Bitwarden Passwordless register form.

    Allows the user to register an account with the application by creating
    and registrering a passkey with the user's device.

    Parameters
    ----------
    client : streamlit_passwordless.BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless application.

    db_session : streamlit_passwordless.db.Session
        An active database session.

    is_admin : bool, default False
        True means that the user will be registered as an admin.
        Not implemented yet.

    pre_authorized : bool, default False
        If True require a user with the input username to exist in the database to allow
        the user to register a new passkey credential. If False omit this validation.

    with_displayname : bool, default False
        If True the displayname field will be added to the form allowing
        the user to fill out a displayname for the account.

    with_alias : bool, default False
        If True the alias field will be added to the form allowing the user to fill
        out one or multiple aliases for the account. Aliases should be separated by
        by semicolon (";"). An alias must be unique across all users.

    title : str, default '#### Register a new passkey with your device'
        The title of the registration from. Markdown is supported.

    border : bool, default True
        If True a border will be rendered around the form.

    validate_button_label : str, default 'Validate'
        The label of the validate button. The validate button validates the fields of the form.

    register_button_label : str, default 'Register'
        The label of the register button. The register button registers a new passkey with the
        user's device. It is enabled when the form is valid.

    validate_button_type : Literal['primary', 'secondary'] or None, default None
        The styling of the validate button. Emulates the `type` parameter of :func:`streamlit.button`.
        If None the button type will be 'secondary' when the form is valid and 'primary' when invalid.

    register_button_type : Literal['primary', 'secondary'], default 'primary'
        The styling of the register button. Emulates the `type` parameter of :func:`streamlit.button`.

    clear_on_validate : bool, default False
        True if the form fields should be cleared when the validate form button is clicked
        and False otherwise.

    Other Parameters
    ----------------
    username_label : str, default 'Username'
        The label of the username field.

    username_max_length : int or None, default 50
        The maximum allowed number of characters of the username field.
        If None the upper limit is removed.

    username_placeholder : str or None, default 'john.doe@example.com'
        The placeholder of the username field. If None the placeholder is removed.

    username_help : str or None, default '__default__'
        The help text to display for the username field. If '__default__' a sensible default
        help text will be used and if None the help text is removed.

    displayname_label : str, default 'Displayname'
        The label of the displayname field.

    displayname_max_length : int or None, default 50
        The maximum allowed number of characters of the displayname field.
        If None the upper limit is removed.

    displayname_placeholder : str or None, default 'John Doe'
        The placeholder of the displayname field. If None the placeholder is removed.

    displayname_help : str or None, default '__default__'
        The help text to display for the displayname field. If '__default__' a sensible default
        help text will be used and if None the help text is removed.

    alias_label : str, default 'Alias'
        The label of the alias field.

    alias_max_length : int or None, default 50
        The maximum allowed number of characters of the alias field.
        If None the upper limit is removed.

    alias_placeholder : str or None, default 'j;john;jd'
        The placeholder of the alias field. If None the placeholder is removed.

    alias_help : str or None, default '__default__'
        The help text to display for the alias field. If '__default__' a sensible default
        help text will be used and if None the help text is removed.
    """

    user = None
    error_msg = ''
    use_default_help = '__default__'
    _help: str | None = None
    banner_container = st.empty()

    with st.container(border=border):
        st.markdown(title)
        with st.form(key=ids.BP_REGISTER_FORM, clear_on_submit=clear_on_validate, border=False):
            if username_help == use_default_help:
                _help = 'A unique identifier for the account. E.g. an email address.'
            else:
                _help = username_help

            username = st.text_input(
                label=username_label,
                placeholder=username_placeholder,
                max_chars=username_max_length,
                help=_help,
                key=ids.BP_REGISTER_FORM_USERNAME_TEXT_INPUT,
            )
            if with_displayname:
                if displayname_help == use_default_help:
                    _help = 'A descriptive name of the user.'
                else:
                    _help = displayname_help

                displayname = st.text_input(
                    label=displayname_label,
                    placeholder=displayname_placeholder,
                    max_chars=displayname_max_length,
                    help=_help,
                    key=ids.BP_REGISTER_FORM_DISPLAYNAME_TEXT_INPUT,
                )
            else:
                displayname = None

            if with_alias:
                if alias_help == use_default_help:
                    _help = (
                        'One or more aliases that can be used to sign in to the account. '
                        'Aliases are separated by semicolon (";"). The username is always '
                        'added as an alias. An alias must be unique across all users.'
                    )
                else:
                    _help = alias_help

                aliases = st.text_input(
                    label=alias_label,
                    placeholder=alias_placeholder,
                    max_chars=alias_max_length,
                    help=_help,
                    key=ids.BP_REGISTER_FORM_ALIASES_TEXT_INPUT,
                )
            else:
                aliases = None

            form_is_valid = st.session_state[config.SK_REGISTER_FORM_IS_VALID]
            if validate_button_type is None:
                button_type: core.ButtonType = 'secondary' if form_is_valid else 'primary'
            else:
                button_type = validate_button_type

            try:
                origin = get_origin_header()
            except exceptions.StreamlitPasswordlessError as e:
                error_msg = str(e)
                origin = ''

            st.form_submit_button(
                label=validate_button_label,
                type=button_type,
                on_click=_validate_form,
                kwargs={
                    'db_session': db_session,
                    'client': client,
                    'origin': origin,
                    'pre_authorized': pre_authorized,
                },
            )

        db_user = st.session_state.get(config.SK_DB_USER)

        disabled = not form_is_valid
        register_token = ''

        if form_is_valid:
            user, error_msg = _create_user(
                username=username,
                user_id=db_user.user_id if db_user else None,
                displayname=displayname,
                aliases=aliases,
            )
            if user is not None:
                register_token, error_msg = _create_register_token(client=client, user=user)

        token, error, clicked = register_button(
            register_token=register_token,
            public_key=client.public_key,
            credential_nickname=username,
            disabled=disabled,
            label=register_button_label,
            button_type=register_button_type,
            key=ids.BP_REGISTER_FORM_SUBMIT_BUTTON,
        )

    if disabled or not clicked:
        return

    if user is None:
        with banner_container:
            st.error(error_msg, icon=config.ICON_ERROR)
        return

    if not token and error:
        error_msg = f'Error creating passkey for user ({username})!\nerror : {error}'
        logger.error(error_msg)
    elif not token:
        error_msg = 'Unexpected error for missing token!'
        logger.error(error_msg)

    if error_msg:
        with banner_container:
            st.error(error_msg, icon=config.ICON_ERROR)
        return

    # The user is still registered even though the sign in may fail!
    verified_user, error_msg = core.verify_sign_in(client=client, token=token)
    if verified_user is None or not verified_user.success:
        error_msg = f'User {username} was registered, but the sign in attempt with registered passkey failed!'
        with banner_container:
            st.error(error_msg, icon=config.ICON_ERROR)

    if not db_user:
        user_create = db.UserCreate(
            user_id=user.user_id, username=user.username, displayname=user.displayname
        )
        db_user = db.create_user(session=db_session, user=user_create)

    msg = f'Successfully registered user: {db_user.username}!'
    logger.info(msg)
    with banner_container:
        st.success(msg, icon=config.ICON_SUCCESS)



r"""The database functions that are cached by Streamlit."""

# Standard library

# Third party
import streamlit as st

# Local
from . import core


@st.cache_resource()
def create_session_factory(
    url: core.URL,
    autoflush: bool = False,
    expire_on_commit: bool = False,
    create_database: bool = True,
    **engine_config,
) -> core.SessionFactory:
    r"""Create the database session factory, which can produce database sessions.

    The database engine is bound to each session produced by the factory.

    Parameters
    ----------
    url : str or sqlalchemy.URL
        The SQLAlchemy database url.

    autoflush : bool, default False
        Automatically flush pending changes within the session to the database
        before executing new SQL statements.

    expire_on_commit : bool, default False
        If True make the connection between the models and the database expire after a
        transaction within a session has been committed and if False make the database models
        accessible after the commit.

    create_database : bool, default True
        If True the database table schema will be created if it does not exist.

    **engine_config : dict
        Additional keyword arguments passed to the :func:`sqlalchemy.create_engine` function.

    Returns
    -------
    session_factory : sqlalchemy.orm.Session
        The session factory that can produce new database sessions.
    """

    session_factory, engine = core.create_session_factory(
        url=url,
        autoflush=autoflush,
        expire_on_commit=expire_on_commit,
        create_database=create_database,
        **engine_config,
    )

    return session_factory



r"""Contains the Streamlit configuration of the web components."""

# Standard library

# Third party
import streamlit as st

# Local


# =====================================================================================
# Session state keys
# =====================================================================================

SK_BP_VERIFIED_USER = 'stp-bp-verified-user'
SK_DB_USER = 'stp-db-user'
SK_REGISTER_FORM_IS_VALID = 'stp-register-form-is-valid'

# =====================================================================================
# Icons
# =====================================================================================

ICON_ERROR = '🚨'
ICON_SUCCESS = '✅'
ICON_WARNING = '⚠️'


# =====================================================================================
# Functions
# =====================================================================================


def init_session_state() -> None:
    r"""Initialize the session state.

    Session state keys
    ------------------
    SK_BP_VERIFIED_USER : streamlit_passwordless.BitwardenPasswordlessVerifiedUser or None, default None
        A verified user from Bitwarden Passwordless.

    SK_DB_USER : streamlit_passwordless.db.models.User or None, default None
        The database user. If None the user does not exist in the database.

    SK_REGISTER_FORM_IS_VALID : bool, default False
        True if the input to register form is valid and False otherwise.
    """

    none_keys = (SK_BP_VERIFIED_USER, SK_DB_USER)
    for key in none_keys:
        st.session_state[key] = None

    st.session_state[SK_REGISTER_FORM_IS_VALID] = False



r"""The database functionality of streamlit-passwordless."""

from . import models
from .cache import create_session_factory
from .core import Session
from .crud.user import create_user, get_all_users, get_user_by_username
from .schemas.user import UserCreate



r"""Functions to work with the browser and the underlying websocket connection of Streamlit."""

# Standard library
import logging

# Third party
import streamlit as st

if (context := getattr(st, 'context', None)) is None:
    from streamlit.web.server.websocket_headers import _get_websocket_headers
else:  # streamlit >= 1.37

    def _get_websocket_headers() -> dict[str, str] | None:
        return getattr(context, 'headers', None)


# Local
from streamlit_passwordless import exceptions

logger = logging.getLogger(__name__)


def get_websocket_headers() -> dict[str, str]:
    r"""Get the HTTP request headers passed along to the websocket connection.

    Returns
    -------
    headers : dict[str, str]
        A copy of the HTTP request headers.

    Raises
    ------
    streamlit_passwordless.StreamlitPasswordlessError
        If no headers could be retrieved from the websocket connection.
    """

    headers = _get_websocket_headers()

    if headers is None:
        error_msg = 'No HTTP request headers received from websocket connection!'
        logger.error(error_msg)
        raise exceptions.StreamlitPasswordlessError(error_msg)

    return headers


def get_origin_header() -> str:
    r"""Get the Origin HTTP request header of to the websocket connection.

    Returns
    -------
    origin : str
        The Origin http request header.

    Raises
    ------
    streamlit_passwordless.StreamlitPasswordlessError
        If no headers could be retrieved from the websocket connection or if
        the "Origin" header was not found.
    """

    headers = get_websocket_headers()
    origin = headers.get('Origin')

    if origin is None:
        error_msg = 'Origin http header not found in websocket connection!'
        logger.error(error_msg)
        raise exceptions.StreamlitPasswordlessError(error_msg)

    return origin



r"""Helper functions and core components that can be used by other components."""

# Standard library
import logging
from typing import Literal, TypeAlias

# Third party
import streamlit as st

# Local
from streamlit_passwordless import exceptions
from streamlit_passwordless.bitwarden_passwordless.backend import BitwardenPasswordlessVerifiedUser
from streamlit_passwordless.bitwarden_passwordless.client import BitwardenPasswordlessClient

from . import config

ButtonType: TypeAlias = Literal['primary', 'secondary']
logger = logging.getLogger(__name__)


def verify_sign_in(
    client: BitwardenPasswordlessClient, token: str
) -> tuple[BitwardenPasswordlessVerifiedUser | None, str]:
    r"""Verify the sign in token with the backend to complete the sign in process.

    Parameters
    ----------
    client : BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless API.

    token : str
        The token to verify.

    Returns
    -------
    verified_user : streamlit_passwordless.BitwardenPasswordlessVerifiedUser or None
        Details from Bitwarden Passwordless about the user that was signed in.
        None is returned if an error occurred during the sign in process. `verified_user` is
        also stored in the session state with the key `config.SK_BP_VERIFIED_USER`.

    error_msg : str
        An error message about what failed during the sign in process.
        An empty string is returned if no errors occurred.
    """

    error_msg = ''
    verified_user: BitwardenPasswordlessVerifiedUser | None = None

    try:
        verified_user = client.verify_sign_in(token=token)
    except exceptions.SignInTokenVerificationError as e:
        error_msg = str(e)
        logger.error(error_msg)
    except exceptions.StreamlitPasswordlessError as e:
        error_msg = f'Error creating verified user!\n{str(e)}'
        logger.error(error_msg)

    st.session_state[config.SK_BP_VERIFIED_USER] = verified_user

    return verified_user, error_msg
